/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "common/common.h"
#include "common/library_fn.h"
#include "common/main.h"

/*Global variables*/
CFG_SECTION_T target_config[MAX_CFG_SECTIONS];
ERROR_MAP_T error_map[] = \
            { \
                {SUCCESS, "Successful operation!"}, \
                {CONFIG_FAILED, "Failed to apply valid configuration!"}, \
                {IO_ERROR, "Failed to perform file IO operation"}, \
                {MALFORMED_CONFIG, "target.config is malformed"}, \
                {NOT_IMPLEMENTED, "Not Implemented"}, \
                {INCOMPLETE_CONFIG, "target.config does not have the expected configuration"}, \
                {UNSUPPORTED_CONFIG, "target.config does have configurations with unsupported values"}, \
                {OS_ERROR, "Fatal OS error"}, \
                {OPERATION_ERROR, "A logical operation failed"} \
            };
char perf_cmdline[512];
char hwmon_cmdline[512];
char multiobj_pmu_ff[512];
char multiobj_hwmon_ff[512];
char report_singleobj_name[32];
char report_shadowobj_name[32];
char report_composite_name[32];
char report_pmu_name[32];
char report_hwmon_name[32];
int perf_max_counters = 0;
int iterator = 0;
int generations = 0;
int restart_count = 0;
struct mq_attr attr;
time_t finish_time;


/*******************************************************************************
 * void display_target_config (void)                                           *
 *    - This method initialises the test suite                                 *
 *    - arg1 - None                                                            *
 ******************************************************************************/
void display_target_config(void)
{

    int i;
    int j;

    REPORT_FN_DETAILS;

    printf("----------Config from target.config----------------\n");
    for (i = 0; i < MAX_CFG_SECTIONS; i++)
    {
        printf("Section[%d] -> [%s]\n",i,target_config[i].section);

        for(j = 0; j < MAX_ITEMS_PER_CFG_SECTION; j++)
        {
            printf("\tConfig[%d] :: %s = %s\n", j, target_config[i].config[j].cfg, target_config[i].config[j].cfg_val);
        }
    }

}


/*******************************************************************************
 * int parse_target_config (void)                                              *
 *    - This method initialises the test suite                                 *
 *    - arg1 - None                                                            *
 ******************************************************************************/
int parse_target_config(void)
{
    FILE* cfg_fd = NULL;
    char str[CFG_STR_LEN];
    char* tmpstr;
    char* freemem;
	char * token;
	int found_section = FALSE;
	int found_start = FALSE;
	int found_end = FALSE;
	int section_index = 0;
	int cfg_index = 0;
	int ret_val = MALFORMED_CONFIG;
	int len = 0;

    REPORT_FN_DETAILS;
    cfg_fd  = fopen("target.config","r");
	if (NULL == cfg_fd)
	{
		printf("MAIN$$ File open error for target.config with error, Terminating...\n");
		return 1;
	}

    /*Read contents from target.config and populate the internal data structure*/
    while (NULL != fgets(str,CFG_STR_LEN,cfg_fd))
    {

        if (section_index < MAX_CFG_SECTIONS)
        {
            tmpstr = strdup(str);
            freemem = tmpstr; // This is needed for freeing the memory allocated by strdup.

//            printf("%x %x\n",tmpstr, freemem);
            if (FALSE == found_section)
            {

                /*Find the section and extract the section name*/
                token = NULL;
     //           printf("%s",tmpstr);
                token = strsep(&tmpstr,"[");
                if (NULL != token)
                {
    //                printf("After tokenising with [ we get $$ %s $$ %s",token,tmpstr);
    //                getchar();
                    if (0 != strcmp(token,str))
                    {
                        len = strlen(tmpstr);
                        tmpstr[len-2] = '\0';
                        strcpy(target_config[section_index].section,tmpstr);
    //                    printf("Got the token!!!! - $$ %s $$\n", tmpstr);
                        found_section = TRUE;
                        found_start = FALSE;
                        found_end = FALSE;
                        cfg_index = 0;
                        ret_val = SUCCESS;
                    }
                }
            }
            else
            {
                /*As we have found the section, let us extract the config elements
                  but first we need to get the start of section token */
    //            printf("In a section now!\n");
                ret_val = MALFORMED_CONFIG;
                if (FALSE == found_start)
                {
    //                printf("%s\n",str);
                    if (0 != strcmp(str,".start\n"))
                    {
                        continue;
                    }
                    else
                    {
          //              printf("Found start of section\n");
                        found_start = TRUE;
                        ret_val = SUCCESS;
                    }

                }/*As we have found the start, let us grab he config elements from
                 the section now*/
                else if ((TRUE == found_start) && (FALSE == found_end))
                {
                    if (0 == strcmp(str,".end\n"))
                    {
          //              printf("Found end of section\n");
                        found_end = TRUE;
                        /* Increment the section index only at this point when an entire section is populated
                        to the internal data structure*/
                        section_index++;
                        found_section = FALSE;
                        ret_val = SUCCESS;
                    }
                    else
                    {
                        if (cfg_index < MAX_ITEMS_PER_CFG_SECTION)
                        {
        //                    printf("%s\n",tmpstr);
                            token = NULL;
                            token = strsep(&tmpstr,":");
                            if (NULL != token)
                            {
        //                        printf("After tokenising with : we get $$ %s $$ from string %s",token,tmpstr);
        //                        getchar();
                                if (0 != strcmp(token,str))
                                {
        //                            printf("Got the token!!!! - $$ %s $$\n", token);
                                    strcpy(target_config[section_index].config[cfg_index].cfg,token);
                                    len = strlen(tmpstr);
                                    tmpstr[len-1] = '\0';
        //                            printf("Got the token!!!! - $$ %s $$\n", tmpstr);
                                    strcpy(target_config[section_index].config[cfg_index].cfg_val,tmpstr);
                                    found_section = TRUE;
                                    cfg_index++;
                                    ret_val = SUCCESS;
                                }
                            }
                        }
                        else
                        {
                            printf("WARNING! Exceeded maximum allowed %d items in a section in target.config, Ignoring and proceeding to the next\n",MAX_ITEMS_PER_CFG_SECTION);
                        }
                    }
                }
            }
            /*Free the memory allocated by strdup. Since tmpstr gets manged by strsep, we use freemem variable for free*/
//            printf("%x %x\n",tmpstr, freemem);
            free(freemem);
        }
        else
        {
            printf("WARNING! Exceeded maximum allowed %d sections in target.config, Ignoring the rest\n",MAX_CFG_SECTIONS);
            break;
        }

    }

    if (MALFORMED_CONFIG == ret_val)
    {
        printf("MAIN$$ ERROR config malformed - %s -\n",error_map[MALFORMED_CONFIG].err_message);
        return MALFORMED_CONFIG;

    }

    return SUCCESS;
}

/*******************************************************************************
 * static int setup_exec_env(void)                                             *
 *    - This method sets up the essential exec env                             *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int setup_exec_env(void)
{
    int ret_val = -1;
    struct stat buf;

    REPORT_FN_DETAILS;

    /*Ensure the fs is mounted in rw mode*/
	if (-1 == system("mount -o rw,remount /"))
    {
	    printf("MAIN$$ Read-Only filesystem, Please make it Read-Write\n");
	    return CONFIG_FAILED;
	}

    /*First determine if /bin/sh exists; if so exec env OK!*/
    if (0 != stat("/bin/sh",&buf))
    {
        printf("MAIN$$ /bin/sh does not exist\n");

        /*Now check if we are in an android env;
          if so create a symlink /bin/sh to /system/bin/sh*/
        if (0 == stat("/system/bin/sh",&buf))
        {
            if (0 == mkdir("/bin", 0777))
            {
                if (0 == symlink("/system/bin/sh", "/bin/sh"))
                {
                    return SUCCESS;
                }
            }
        }

        /*If the control comes here means it is exec env could not be setup!*/
        printf("MAIN$$ Failed creating exec env!\n");
        return CONFIG_FAILED;
    }
    else
        return SUCCESS;

}

/*******************************************************************************
 * int initialise_testsuite(FIT_FN_T* fitnessfn, PL_T* payload, ALG_T* algo)   *
 *    - This method initialises the test suite                                 *
 *    - arg1 - fitnessfn - The output variable that would indicate the         *
 *             fitness funtion type as discovered from the config file         *
 *    - arg2 - payload - The output variable that would indicate the payload   *
 *             type as discovered from the config file                         *
 *    - arg3 - algo - The output variable that would indicate the algorithm    *
 *             type for search state exploration as discovered from the config *
 *             file                                                            *
 ******************************************************************************/
int initialise_testsuite(FIT_FN_T* fitnessfn, PL_T* payload, ALG_T* algo)
{
    int ret_val = -1;

    REPORT_FN_DETAILS;

    printf("MAIN$$ Initialising test suite...!\n");

    ret_val = setup_exec_env();
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - %s -\n",error_map[ret_val].err_message);

        return CONFIG_FAILED;
    }

    ret_val = parse_target_config();
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - %s -\n",error_map[ret_val].err_message);

        return CONFIG_FAILED;
    }

    display_target_config();

    ret_val = fitness_function_init_config(fitnessfn);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Fitness Function config initialisation failed with %s -\n",error_map[ret_val].err_message);

        return CONFIG_FAILED;
    }

    ret_val = payload_init_config(payload);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Payload config initialisation failed with %s -\n",error_map[ret_val].err_message);

        return CONFIG_FAILED;
    }

    ret_val = sbst_algo_init_config(algo);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - SBST algorithm config initialisation failed with %s -\n",error_map[ret_val].err_message);

        return CONFIG_FAILED;
    }

    /*Setup the common message Q attributes*/
 	attr.mq_flags = 0; /* i.e mq_send will be block if message queue is full */
 	attr.mq_maxmsg = MQ_MAX_NUM_OF_MESSAGES;
 	attr.mq_msgsize = MQ_MESSAGE_MAX_LENGTH;
 	attr.mq_curmsgs = 0; /* mq_curmsgs is don't care */

    return SUCCESS;
}

/*******************************************************************************
 * int main()                                                                  *
 *    - This is the main entry function in this program                        *
 ******************************************************************************/
int main()
{
    int ret_val = -1;
    FIT_FN_T ff = INVALID_FF;
    PL_T pl = INVALID_PL;
    ALG_T alg = INVALID_ALG;
	long  hours = 0;
	long minutes = 0;
	long seconds = 0;
	double execution_time = 0;
	time_t start;
	time_t end;
	time_t restart_end;
	char str[128];
	FILE* report_pmu_fd = NULL;
	FILE* report_hwmon_fd = NULL;
	FILE* report_composite_fd = NULL;


    REPORT_FN_DETAILS;

    time(&start); /*Start clocking the experiment*/
    finish_time = time(NULL) + EXPERIMENT_DURATION_IN_SEC;

    printf("MAIN$$ OnTartget execution initiated...!\n");
    /*Initialise the test suite based on the parameters parsed from the target.config file*/
    ret_val = initialise_testsuite(&ff, &pl, &alg);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Test Suite Initialisation Failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        exit(0); //We terminate here prematurely. Fatal Error encountered!
    }

    /*At this point we have primed the test suite for execution with a successful init
      Now run the suite.*/
    /*Fitness Function startup so the fitness function thread is launched and listening for commands*/
    ret_val = fitness_function_startup(ff);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Fitness Function Startup failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        goto FF_TEARDOWN;
    }


    /*Followed by payload startup - In our case this encapsulates the search space which launches
      the payload thread and listening for commands*/
    ret_val = payload_startup(pl);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Payload Startup failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        goto PL_TEARDOWN;
    }

      /*Run the SBST algorithm of choice
      First initialise the state space and seed the search*/
    ret_val = sbst_algo_init_statespace_and_randomseed(alg, pl, ff);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - SBST algorithm initialisation fo state space and random seeding failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        goto ALG_TEARDOWN;
    }

    ret_val = sbst_algo_iterate_statespace(alg, pl, ff);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - SBST algorithm iterating state space search failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        /*Since the iterate state space failed proceed with orderly teardown sequence*/
    }


    /*Wrap-up the execution - last init to be teardown first!*/
    ALG_TEARDOWN: //This is the lable for goto method to ennsure right teardown sequence

    ret_val = sbst_algo_teardown(alg, pl, ff);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - SBST algorithm teardown failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Terminating Execution!!!\n");
        /*We should let remaining teardown operations to happen before we give up*/
    }

//    printf("Hit any key - After ALG teardown");
//    getchar();

    PL_TEARDOWN: //This is the label for goto method to ensure right teardown sequence

    ret_val = fitness_function_teardown(ff);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Fitness Function Teardown failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Attempt further teardown before Terminating Execution!!!\n");
        /*We should let remaining teardown operations to happen before we give up*/
    }

//    printf("Hit any key - After PL teardown");
//    getchar();

    FF_TEARDOWN: //This is the label for goto method to ensure right teardown sequence
        /*If the ff init sequence prematurely failed invoke only that tear down sequence for
          any possible cleanup actions!*/
    ret_val = payload_teardown(pl);
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR - Payload Teardown failed with error %d -\n",error_map[ret_val].err_code);
        printf("MAIN$$ Attempt further teardown before Terminating Execution!!!\n");
        /*We should let remaining teardown operations to happen before we give up*/
    }

//    printf("Hit any key - After FF teardown");
//    getchar();

  	time(&end); /*Clocking the experiment duration*/

    execution_time = difftime(end,start);

    hours = ((long)execution_time % SEC_IN_DAY) / SEC_IN_HOUR;
    minutes = ((long)execution_time % SEC_IN_HOUR) / SEC_IN_MINUTE;
    seconds = (long)execution_time % SEC_IN_MINUTE;

	printf("MAIN$$ Total time = %f hours = %ld minutes = %ld seconds = %ld\n",execution_time,hours,minutes,seconds);

    /*Log the experiment execution time into the report*/
//    report_fd = fopen("report.txt","a+");
//    if (NULL == report_fd)
//    {
//        printf("MAIN$$ File open error for report.txt, Terminating...\n");
//        return IO_ERROR;
//    }

//	sprintf(str,"\n#################################################################\nTotal Execution took %ld:%ld:%ld (hh:mm:ss)\n",hours,minutes,seconds);

//	fputs(str,report_fd);

//    fclose(report_fd);

    report_pmu_fd = fopen(report_pmu_name,"a+");
    if (NULL == report_pmu_fd)
    {
        printf("MAIN$$ File open error for %s\n", report_pmu_name);
        //Just report the error as the report_pmu.txt is not created in this instance.
    }
    else
    {
    	sprintf(str,"\n#################################################################\nTotal Execution took %ld:%ld:%ld (hh:mm:ss)\n",hours,minutes,seconds);

    	fputs(str,report_pmu_fd);

        fclose(report_pmu_fd);

    }


    report_hwmon_fd = fopen(report_hwmon_name,"a+");
    if (NULL == report_hwmon_fd)
    {
        printf("MAIN$$ File open error for %s\n", report_hwmon_name);
        //Just report the error as the report_hwmon.txt is not created in this instance.
    }
    else
    {
    	sprintf(str,"\n#################################################################\nTotal Execution took %ld:%ld:%ld (hh:mm:ss)\n",hours,minutes,seconds);

	    fputs(str,report_hwmon_fd);

        fclose(report_hwmon_fd);

    }


    report_composite_fd = fopen(report_composite_name,"a+");
    if (NULL == report_composite_fd)
    {
        printf("MAIN$$ File open error for %s\n", report_composite_name);
        //Just report the error as the report_hwmon.txt is not created in this instance.
    }
    else
    {
    	sprintf(str,"\n#################################################################\nTotal Execution took %ld:%ld:%ld (hh:mm:ss)\n",hours,minutes,seconds);

	    fputs(str,report_composite_fd);

        fclose(report_composite_fd);

    }

    return SUCCESS;
}