/********************************************************************************
 *  Copyright 2018-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/multi_objective_fitness_function.h"

/*******************************************************************************
 * int report_multi_obj_fitness_function (int node_index)                            *
 *    - This function evaluates the pmu based fitness function using perf      *
 *      based pmu counters                                                     *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_multiobj_fitness_function(int node_index)
{

    int ret_val1 = -1;
    int ret_val2 = -1;

    REPORT_FN_DETAILS;

	ret_val1 = report_pop_based_pmu_fitness_function(node_index);
    if (SUCCESS != ret_val1)
	{
		printf("MAIN$$ ERROR: Unable to report pmu fitness function from multi-objective fitness function\n");

	    return OPERATION_ERROR;
	}

	ret_val2 = report_pop_based_hwmon_fitness_function(node_index);
    if (SUCCESS != ret_val2)
	{
		printf("MAIN$$ ERROR: Unable to report hwmon fitness function from multi-objective fitness function\n");

	    return OPERATION_ERROR;
	}

    return SUCCESS;

}

/*******************************************************************************
 * int launch_multiobj_fitness_function(void)                                  *
 *    - This function creates a thread to evaluate the fitness function        *
 *      using perf based pmu counters                                          *
 ******************************************************************************/

int launch_multiobj_fitness_function(void)
{

	int ret_val1 = -1;
	int ret_val2 = -1;

	REPORT_FN_DETAILS;

    ret_val1 = launch_pmu_fitness_function();
    if (ret_val1)
	{
		printf("MAIN$$ ERROR: Unable to create pmu fitness function thread as part of multi-objective fitness function\n");

	    return OS_ERROR;
	}

    ret_val2 = launch_hwmon_fitness_function();
    if (ret_val2)
	{
		printf("MAIN$$ ERROR: Unable to create hwmon fitness function thread as part of multi-objective fitness function\n");

	    return OS_ERROR;
	}

	return SUCCESS;

}

/*******************************************************************************
 * int report_composite_fitness_function (int node_index)                      *
 *    - This function evaluates the composite fitness function using normalised*
 *      pmu scores and hwmon values                                            *
 *    - arg1 - node_index - specifies the node operated on                     *
 ******************************************************************************/
int report_composite_fitness_function(int node_index)
{

    int ret_val1 = -1;
    int ret_val2 = -1;
    int normalised_pmu_score = 0;
    int normalised_hwmon_score = 0;
    int pmu_lower_scale = 0;
    int hwmon_lower_scale = 0;
    /*upperscale = average + 3 * SD*/
    int pmu_upper_scale = 2017;
    int hwmon_upper_scale = 736;

    REPORT_FN_DETAILS;



    /*Normalise the scores to a range of 1 (mapped from A) to 100 (mapped from B)
      y = 1 + (x-A)*(100-1)/(B-A)
    */
    if (-1 == node_index)
    {
        normalised_pmu_score = 1 + (current_node.pmu_computed_fitness_score.counter_score[0] - pmu_lower_scale) * \
                                (100 - 1) / (pmu_upper_scale - pmu_lower_scale);

        normalised_hwmon_score = 1 + (current_node.hwmon_computed_fitness_score.counter_score[0] - hwmon_lower_scale) * \
                                (100 - 1) / (hwmon_upper_scale - hwmon_lower_scale);

        current_node.composite_fitness_score.pmu_normalised = normalised_pmu_score;
        current_node.composite_fitness_score.hwmon_normalised = normalised_hwmon_score;
    }
    else if ((node_index >= 0) && (node_index < NEXT_GEN_CANDIDATE_COUNT))
    {
        normalised_pmu_score = 1 + (next_gen[node_index].pmu_computed_fitness_score.counter_score[0] - pmu_lower_scale) * \
                                (100 - 1) / (pmu_upper_scale - pmu_lower_scale);

        normalised_hwmon_score = 1 + (next_gen[node_index].hwmon_computed_fitness_score.counter_score[0] - hwmon_lower_scale) * \
                                (100 - 1) / (hwmon_upper_scale - hwmon_lower_scale);

        next_gen[node_index].composite_fitness_score.pmu_normalised = normalised_pmu_score;
        next_gen[node_index].composite_fitness_score.hwmon_normalised = normalised_hwmon_score;

    }
    else
	    printf("MAIN$$ Should not reach here.....!!!!\n");

    return SUCCESS;

}