#    Copyright 2016-2020 University of Leicester & Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABCMeta, abstractmethod
from sbstauto import config


class perfCounter():

    __metaclass__ = ABCMeta

    @abstractmethod
    def listEventCounters(self):
        """
        This method shows the events set to be counted with perf utility
        :return: None
        """
        pass

    @abstractmethod
    def configEventCounters(self, event1, *args):
        """
        This config method sets the events to be counted with perf utility
        :param counter1: first event to be monitored
        :param args: additional events to be monitored
        :return: None if successfull;
        raises an OverflowError if more events than supported is specified;
        raises a ValueError exception if unsupported event is specified
        """
        pass

    @abstractmethod
    def _addToEventList(self, event):
        """
        This internal method adds the events to be monitored to the event list
        after checking if it is supported event for monitoring
        :param event: The event to be monitored
        :return: None if successfull;
        raises an OverflowError if more events than supported is specified;
        raises a ValueError exception if unsupported event is specified
        """
        pass

    @abstractmethod
    def resetEventList(self):
        """
        This method resets the events list to empty list
        :return: None
        """
        pass

    @abstractmethod
    def finaliseConfig(self):
        """
        This method writes the perf command line based on the events selected
        for monitoring during the execution
        :return: None if successfull;
        raises a OSError exception if config could not be updated
        raises a ValueError exception if the eventlist was empty
        """
        pass

    def _generatePerfCmdLine(self):
        """
        This internal method generates the perf command line string
        based on the added events
        :return: the commandline string
        """

class TC2perfCounter(perfCounter):

    _MAXCOUNTERS = 4
    _supportedevent = set(["cci_a7_read_data_stall", "cci_a15_read_data_stall",
                           "cci_a7_write_data_stall_trackerfull",
                           "cci_a15_write_data_stall_trackerfull",
                           "cci_a7_read_data_stall_trackerfull",
                           "cci_a15_read_data_stall_trackerfull",
                           "cci_a7_write_data_stall", "cci_a15_write_data_stall",
                           "cci_a15_read_req_stall_id_hzrd",
                           "cci_a7_read_req_stall_id_hzrd"])
    _eventmap = {"cci_a7_read_data_stall": "0x8b", "cci_a15_read_data_stall": "0x6b",
                 "cci_a7_write_data_stall_trackerfull": "0x93",
                 "cci_a15_write_data_stall_trackerfull": "0x73",
                 "cci_a7_read_data_stall_trackerfull": "0x89",
                 "cci_a15_read_data_stall_trackerfull": "0x69",
                 "cci_a7_write_data_stall": "0x93", "cci_a15_write_data_stall": "0x73",
                 "cci_a15_read_req_stall_id_hzrd": "0x74",
                 "cci_a7_read_req_stall_id_hzrd": "0x94"}
    _perfexecutable = config.SBSTAUTO_HOME+"/tools/external/armeabi/perf"

    def __init__(self):
        self._eventlist = []


    def listEventCounters(self):
        for item in self._eventlist:
            print item


    def configEventCounters(self, event1, *args):
        try:
            self._addToEventList(event1)
            for arg in args:
                self._addToEventList(arg)
        except (ValueError, OverflowError) as e:
            print("{}".format(str(e)))
            raise e

    def _addToEventList(self, event):
        if len(self._eventlist) < self._MAXCOUNTERS:
            if event in self._supportedevent:
                self._eventlist.append(event)
            else:
                raise ValueError("PMUCOUNTER ERROR: {} is not a supported event".format(event))
        else:
            raise OverflowError("PMUCOUNTER ERROR: {} could not be added as maximum of {} "
                                "events already added".format(event,self._MAXCOUNTERS))

    def resetEventList(self):
        self._eventlist = []

    def finaliseConfig(self):
        if not self._eventlist:
            raise ValueError("PMUCOUNTER ERROR: No Events selected for monitoring")
        else:
            try:
                perfcmdfound = False
                perfmaxcntrfound = False
                with open("{}/output/target.config.tmpl".format(config.SBSTAUTO_HOME),
                          mode="rt") as f_in, \
                        open("{}/output/target.config".format(config.SBSTAUTO_HOME),
                             mode="wt") as f_out:
                    for line in iter(f_in):
                        if -1 != line.find("perf_cmdline"):
                            perfcmdfound = True
                            cmdline = self._generatePerfCmdLine()
                            f_out.write("perf_cmdline={}".format("\""+cmdline+"\"\n"))
                        if -1 != line.find("max_counters")
                            perfmaxcntrfound = True
                            f_out.write("max_counters={}".format(self._MAXCOUNTERS))
                        else:
                            if perfcmdfound and perfmaxcntrfound and ".end\n" == line:
                                f_out.write(line)
                                return
                            else:
                                f_out.write(line)

            except IOError as e:
                print("PMUCOUNTER ERROR: {}".format(str(e)))
                raise e

    def whichPerfExe(self):
        return self._perfexecutable

    def _generatePerfCmdLine(self):
        cmdlinestring = ""
        for item in self._eventlist:
            try:
                token = "CCI/config={},name={}/,".format(self._eventmap[item], item)
                cmdlinestring = cmdlinestring + token
            except KeyError as e:
                print("PMUCOUNTER_ERROR: {} --> key used is {}".format(str(e), item))
                raise e

        cmdlinestring = cmdlinestring[:-1]
        cmdlinestring = cmdlinestring + " "

        return cmdlinestring

if __name__ == "__main__":
    print("Instantiating TC2perfCounter concrete class")
    pc = TC2perfCounter()
    try:
        pc.listEventCounters()
        pc.configEventCounters("cci_a7_read_data_stall", "cci_a15_read_data_stall",
                     "cci_a7_read_data_stall_trackerfull",
                     "cci_a15_read_data_stall_trackerfull")
        pc.finaliseConfig()
    except (ValueError, OverflowError, IOError) as e:
        print(str(e))
    pc.listEventCounters()



