/********************************************************************************
 *  Copyright 2016-2020 University of Leicester & Arm Limited                   *
 *                                                                              *
 *  Licensed under the Apache License, Version 2.0 (the "License");             *
 *  you may not use this file except in compliance with the License.            *
 *  You may obtain a copy of the License at                                     *
 *                                                                              *
 *     http://www.apache.org/licenses/LICENSE-2.0                               *
 *                                                                              *
 *  Unless required by applicable law or agreed to in writing, software         *
 *  distributed under the License is distributed on an "AS IS" BASIS,           *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    *
 *  See the License for the specific language governing permissions and         *
 *  limitations under the License.                                              *
 *                                                                              *
 *******************************************************************************/

#include "include/payload_cci_mem_operations.h"

/*Global variables*/
PAYLOAD_T current_node;
PAYLOAD_T best_node;
PAYLOAD_T next_gen[NEXT_GEN_CANDIDATE_COUNT];
PAYLOAD_T pop[SPEA2_POPULATION_SIZE+SPEA2_ARCHIVE_SIZE];
PAYLOAD_T pop_nextgen[SPEA2_POPULATION_SIZE+SPEA2_ARCHIVE_SIZE];
PAYLOAD_T pop_matingpool[SPEA2_MATINGPOOL_SIZE];
GRID_POS_T cci_mem_rdwr_payload[PL_SIZE_512M];
int heap_init_size = 0;
/*The rows and cols indicate the dimension of the heap size allocated
  for the experiment*/
int rows = 0;
int cols = 0;
/*The pl_<blah> captures the details of the payload instance dimensions
  used during the specific iteration in the experiment*/
long long pl_size = 0;
int pl_rows_max = 0;
int pl_cols_max = 0;
char read_cpupart[16];
int* array_2d = NULL;
CPU_TOP_T cpu_topology;
THREADDATA_T idwr[MAX_CPU_COUNT];
THREADDATA_T idrd[MAX_CPU_COUNT];
pthread_t data_gen_thread[MAX_CPU_COUNT];
pthread_t data_obsr_thread[MAX_CPU_COUNT];
int dg_thread_count = 0;
int do_thread_count = 0;
mqd_t mqd_wr[MAX_CPU_COUNT];
mqd_t mqd_rd[MAX_CPU_COUNT];

/*******************************************************************************
 * void set_barriers(void)                                                     *
 *    - This function is to set barriers in an attempt to flush any pending    *
 *      cache operations with a view to provide similar start conditions       *
 *      between two iterations                                                 *
 *    - arg1 - none                                                            *
 ******************************************************************************/
void  set_barriers(void)
{

	dmb();
	dsb();

}


/*******************************************************************************
 * int initialise_memory_pattern(void)                                         *
 *    - This function is to set an initial memory pattern after allocating it  *
 *      in the first instance                                                  *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int initialise_memory_pattern(void)
{

    long i = 0;
    long j = 0;
	static int done = 0;

    cols = DEFAULT_COLS;
    rows = (heap_init_size*1024*1024)/(sizeof(int)*cols);

    /*If this is the first instance then allocate the heap memory*/
	if (!done)
	{
        printf("MAIN$$  Rows:: %d Cols:: %d\n",rows,cols);
    	printf("MAIN$$ Allocating memory for first time SINGLE CHUNK\n");
	    array_2d = malloc(heap_init_size*1024*1024);

	    if (NULL == array_2d)
	    {
    	    printf("MAIN$$ Memory allocation failed, Terminating...\n");
    	    return OS_ERROR;
	    }
        done = 1;
	}
    /*Now initialise the memory with a known pattern*/
	printf("MAIN $$ Walkthrough memory SINGLE CHUNK\n");
	printf("0x%x\n",array_2d);


        for (i = 0; i < rows; i++)
        {
                for (j = 0; j < cols; j++)
                {
                        *((int*)array_2d+(cols*i)+j)=(64+j)%64;
                }
        }

        return SUCCESS;
}

/*******************************************************************************
 * void display_memory_pattern(void)                                           *
 *    - This function displays the pattern written on the memory  user as      *
 *    - arg1 - none                                                            *
 ******************************************************************************/
void display_memory_pattern(void)
{

	int i = 0;
	int j = 0;

	for (i = 0; i < rows; i++)
	{
		printf("\n[%04d]:: ",i);
		for (j = 0; j < cols; j++)
		{
			printf("%x",(int*)array_2d+(cols*i)+j);
		}
	}

	return SUCCESS;
}


/*******************************************************************************
 * int reset_to_init_state(void)                                               *
 *    - This function reinitialises the board to the initial data pattern      *
 *      state from which the nextGen operation need to be performed            *
 *    - arg1 - none                                                            *
 ******************************************************************************/
int reset_to_init_state(void)
{

    REPORT_FN_DETAILS;

	/*By performing the same initialisation sequences the expectation is
      that the data cache state at a similar init state if not exact*/

	initialise_memory_pattern();
//	display_memory_pattern();

    return SUCCESS;
}

/*******************************************************************************
 * int display_topology_info(void)                                             *
 *    - This function displays the topology info as populated from the         *
 *      discovery step.                                                        *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int display_topology_info(void)
{
    int i = 0;

    REPORT_FN_DETAILS;

	/*Print the discovered topology info including cluster info and read CPU info*/
    printf("MAIN$$ CPU Topology discovered is -> \n");
    for (i=0; i<cpu_topology.cpu_count; i++)
    {
        printf("CPU %d -> Cluster %d", cpu_topology.cpu_info[i].cpu_id, cpu_topology.cpu_info[i].cluster_id);
        if (cpu_topology.cpu_info[i].is_readcpu)
            printf(" --> READ CPU\n");
        else
            printf("\n");
    }
    printf("\n");

    return SUCCESS;
}


/*******************************************************************************
 * int discover_topology_info(void)                                            *
 *    - This function extracts the on target information about the topology    *
 *      as needed to determine the bumber of threads and their affinity        *
 *    - arg1 - none                                                            *
 ******************************************************************************/
static int discover_topology_info(void)
{
	int ret_val = -1;
	FILE *sysfs_fp;
	char str[SYSFS_STR_LEN];
	char sysfs_path[SYSFS_STR_LEN];
	char * token;
	char* tmpstr01;
	char* freemem01;
	char* tmpstr02;
	char* freemem02;
	int index = 0;
	int lower = 0;
	int higher = 0;
	int i = 0;
	int j = 0;
	int k = 0;
	char siblings[MAX_CLUSTER_COUNT][32];
	int new_cluster = FALSE;
	int matched = FALSE;
	int value = 0;
	char value_str[16];

    REPORT_FN_DETAILS;

    memset(&cpu_topology,0,sizeof(CPU_TOP_T));

  sysfs_fp = fopen("/sys/devices/system/cpu/online","rb");
//	sysfs_fp =  fopen("./online","rb");
	if (NULL == sysfs_fp)
	{
		printf("MAIN$$ ERROR: Discovery of the topology failed. Terminating...\n");
		return OS_ERROR;
	}

    if (NULL != fgets(str,SYSFS_STR_LEN,sysfs_fp))
    {
        printf("MAIN$$ Online CPUs are %s\n",str);
        tmpstr01 = strdup(str);
        freemem01 = tmpstr01;
        while (NULL != tmpstr01)
        {

            if (index >= MAX_CPU_COUNT)
            {
                printf("MAIN$$ Exceeds allowed number of CPUs, Terminating...");

                return OPERATION_ERROR;
            }
            token = NULL;
            /*This one will get the first token which is comma-separated
              that could be either a single number or a hyphen delimited range*/
            token = strsep(&tmpstr01,",");

            tmpstr02 = strdup(token);
            freemem02 = tmpstr02;

            token = NULL;
            token = strsep(&tmpstr02,"-");
            if (NULL == tmpstr02)
            {
                /*If tmpstr02 is NULL implies the token was *not* a range but a single number*/
//                printf("%s :: ",token);
                sscanf(token, "%d", &(cpu_topology.cpu_info[index].cpu_id));
                index++;
                cpu_topology.cpu_count++;
            }
            else
            {
//                printf("%s till %s ::\n",token, tmpstr02);
                sscanf(token, "%d", &lower);
                sscanf(tmpstr02, "%d", &higher);
                for (i=lower; i<=higher; i++)
                {
                    cpu_topology.cpu_info[index].cpu_id = i;
                    index++;
                    cpu_topology.cpu_count++;
                }

            }
            free(freemem02);
        }
        free(freemem01);

    }

    index = 0;
    /*At this point you would have discovered the number of online CPUs and their cpu numbers
      Now go through each CPU and look at core_siblings_list to determine the distinct
      clusters the CPUs are arranged in*/
    for (i=0; i<cpu_topology.cpu_count; i++)
    {
        sprintf(sysfs_path,"/sys/devices/system/cpu/cpu%d/topology/core_siblings_list",cpu_topology.cpu_info[i].cpu_id);
//        sprintf(sysfs_path,"./cpu%d/core_siblings_list",cpu_topology.cpu_info[i].cpu_id);

	    sysfs_fp =  fopen(sysfs_path,"rb");
	    if (NULL == sysfs_fp)
	    {
		    printf("MAIN$$ ERROR: Discovery of the topology failed. Terminating...\n");
		    return OS_ERROR;
	    }

        if (NULL != fgets(str,SYSFS_STR_LEN,sysfs_fp))
        {

            printf("MAIN$$ cpu%d siblings are %s\n",cpu_topology.cpu_info[i].cpu_id,str);

            if (index >= MAX_CLUSTER_COUNT)
            {
                printf("MAIN$$ Exceeds allowed number of Clusters, Terminating...");

                return OPERATION_ERROR;
            }

            new_cluster = FALSE;

            /*First determine if this is new cluster info or not*/
            j = cpu_topology.cluster_count;

            if (0 == j)
            {
                /*This is the first item so add this straight*/
                strcpy(siblings[cpu_topology.cluster_count], str);
                printf("MAIN$$ New cluster found is:: %s\n", siblings[cpu_topology.cluster_count]);
                cpu_topology.cluster_count++;
                new_cluster = TRUE;
            }
            else
            {
                matched = FALSE;
                /*Now iterate across earlier discovered cluster info if we have a new one or not*/
                while (j > 0)
                {
//                    printf("MAIN$$ Comparing cluster info %s with %s\n", str, siblings[j-1]);
                    if (0 == strcmp(siblings[j-1], str))
                    {
                        /*This is an already discovered cluster, so break the loop as
                          nothing more to search*/
                          matched = TRUE;
                          printf("MAIN$$ breaking the loop\n");
                          break;
                    }
                    else
                    {
                        /*Continue to search*/
                        j--;
                    }
                }

                /*A non-zero index for j indicate that this is a new cluster*/
                if (FALSE == matched)
                {
                    strcpy(siblings[cpu_topology.cluster_count], str);
                    printf("MAIN$$ New cluster found is:: %s\n", siblings[cpu_topology.cluster_count]);
                    cpu_topology.cluster_count++;
                    new_cluster = TRUE;
                }

            }

            /*In case of new cluster info assign the cluster numbers to the respective cpus
              in the cpu_info data structure*/
            if (TRUE == new_cluster)
            {

                token = NULL;
                tmpstr01 = strdup(str);
                freemem01 = tmpstr01;
                while (NULL != tmpstr01)
                {
                    /*This one will get the first token which is comma-separated
                      that could be either a single number or a hyphen delimited range*/
                    token = strsep(&tmpstr01,",");

                    tmpstr02 = strdup(token);
                    freemem02 = tmpstr02;

                    token = NULL;
                    token = strsep(&tmpstr02,"-");
                    if (NULL == tmpstr02)
                    {
                        /*If tmpstr02 is NULL implies the token was *not* a range but a single number*/
//                        printf("%s :: ",token);
                        sscanf(token, "%d", &value);
                        /*Now record the cluster info in the cpu_info data structure*/
                        for (j=0; j<cpu_topology.cpu_count; j++)
                        {
//                            printf("%d $$ %d\n",cpu_topology.cpu_info[j].cpu_id, value);
                            if (cpu_topology.cpu_info[j].cpu_id == value)
                            {
                                cpu_topology.cpu_info[j].cluster_id = index;
                                break;
                            }
                        }
                    }
                    else
                    {
//                        printf("%s till %s ::\n",token, tmpstr02);
                        sscanf(token, "%d", &lower);
                        sscanf(tmpstr02, "%d", &higher);
                        for (k=lower; k<=higher; k++)
                        {
                            /*Now record the cluster info in the cpu_info data structure*/
                            for (j=0; j<cpu_topology.cpu_count; j++)
                            {
//                                printf("%d $$ %d\n",cpu_topology.cpu_info[j].cpu_id, k);
                                if (cpu_topology.cpu_info[j].cpu_id == k)
                                {
                                    cpu_topology.cpu_info[j].cluster_id = index;
                                    break;
                                }
                            }
                        }

                    }
                    free(freemem02);
                }
                free(freemem01);
                /*Increment the cluster count index*/
                index++;

            }

        }
	}
    /*If the read_cpupart optional argument is specified capture that info else
      make cluster0 default as read cpu*/
    if (0 == strcmp(read_cpupart,"not-specified"))
    {
        /*Make cluster 0 cpus for data observer threads*/
        for (i=0; i< cpu_topology.cpu_count; i++)
        {
            if (0 == cpu_topology.cpu_info[i].cluster_id)
            {
                cpu_topology.cpu_info[i].is_readcpu = TRUE;
            }
        }
    }
    else
    {
        /*Parse the /proc/cpuinfo to determine the cpus/clusters with readcpu part number*/
      sysfs_fp = fopen("/proc/cpuinfo","rb");
//	    sysfs_fp =  fopen("./cpuinfo","rb");
	    if (NULL == sysfs_fp)
	    {
		    printf("MAIN$$ ERROR: Discovery of the topology failed. Terminating...\n");
		    return OS_ERROR;
	    }

        while (NULL != fgets(str,SYSFS_STR_LEN,sysfs_fp))
        {
//            printf("MAIN$$ Proc Info ---> %s\n",str);

            tmpstr01 = strdup(str);
            freemem01 = tmpstr01;
            token = NULL;
            /*This one will get the first token which is comma-separated
              that could be either a single number or a hyphen delimited range*/
            token = strsep(&tmpstr01,":");
//            printf("Tokenised %s : %s\n",token, tmpstr01);
            if (NULL != strstr(token,"processor"))
            {
                sscanf(tmpstr01,"%d",&value);
//                printf("CPU %d --> ",value);
            }
            else if (NULL != strstr(token,"CPU part"))
            {
                sscanf(tmpstr01,"%s",value_str);
                if (NULL != strstr(value_str,read_cpupart))
                {
//                    printf("%s\n",read_cpupart);

                    /*Find the cpu based in the id and mark for data observer threads*/
                    for (i=0; i< cpu_topology.cpu_count; i++)
                    {
                        if (value == cpu_topology.cpu_info[i].cpu_id)
                        {
                            cpu_topology.cpu_info[i].is_readcpu = TRUE;
                            break;
                        }
                    }

                }


            }

            free(freemem01);

        }
    }

    return SUCCESS;
}

/*----- The following relates to data generator functions ---------------------*/

/*******************************************************************************
 * static void data_operate(TRFM_T type, int cpu_num)                          *
 *    - This function performs some data operations on the memory region       *
 *    - arg1 - type - specifies the type of data operation to be performed     *
 *    - arg2 - cpu_num - identifies the cpu where the data generator thread is *
 *                  running                                                    *
 ******************************************************************************/

static void data_operate (TRFM_T type, int cpu_num)
{

    long long i = 0;
    int j = 0;
	int temp = 0;
	int val = 0;
	long long payload_size = 0;

    REPORT_FN_DETAILS;


	switch (type)
	{

                case PAYLOAD_SIZE:
                            printf("DATAGEN_THREAD[CPU-%d]## !!!!!!!Before PAYLOAD_SIZE_WR %lld !!!!!!!!!\n", cpu_num, pl_size);
    				        payload_size = pl_size;

			                for (i = 0; i < payload_size; i++)
        			        {

                                        *((int*)array_2d+(cols*cci_mem_rdwr_payload[i].x_pos)+cci_mem_rdwr_payload[i].y_pos)=255;
                			}
                           printf("DATAGEN_THREAD[CPU-%d]## !!!!!!!After PAYLOAD_SIZE_WR %lld !!!!!!!!!\n", cpu_num, pl_size);
        		break;

        		default:
		                	/*no-op*/
			                printf("DATAGEN_THREAD[CPU-%d]## Should never reach\n", cpu_num);
			    break;
	}


}


/*******************************************************************************
 * static void * data_generator_main(void)                                     *
 *    - This is the main entry function of the thread that performs the        *
 *      data write                                                             *
 *    - no arguments                                                           *
 ******************************************************************************/

static void * data_generator_main (void * ptr)
{
    mqd_t mqd;
    int ret = -1;
    int i = 0;
    int threaddata = 0;
    char msg_buff[MQ_MESSAGE_MAX_LENGTH];
    ssize_t msg_len;
    int ret_val = -1;
    int finished=0;
	static int dg_counter=0;
	cpu_set_t cpuset;
	pthread_t thread;
 	THREADDATA_T * id;
 	char mq_str[16];

    REPORT_FN_DETAILS;

    id = (THREADDATA_T *)ptr;
    threaddata = id->id;
    strcpy(mq_str, id->mq_name);

    ((THREADDATA_T*)ptr)->threadid = syscall(SYS_gettid);

    printf("DATAGEN_THREAD[CPU-%d]## %d %s\n", threaddata, ((THREADDATA_T*)ptr)->id, ((THREADDATA_T*)ptr)->mq_name);

    printf ("DATAGEN_THREAD[CPU-%d]## In dg_threadid:: %d parent pid:: %d\n", threaddata, ((THREADDATA_T*)ptr)->threadid, getpid());

	CPU_ZERO(&cpuset);

	/*Set the affinity of data generator thread to CPUID passed via threaddata*/
	CPU_SET(threaddata,&cpuset);

	thread = pthread_self();


	ret = pthread_setaffinity_np(thread,sizeof(cpu_set_t),&cpuset);
	if (ret != 0)
	{
		printf("DATAGEN_THREAD[CPU-%d]## @@@@@@@@@@@@@@@@@@FAILED TO SET THREAD AFFINITY FOR OBSERVER THREAD@@@@@@@@@@@@@\n", threaddata);

//        ret_val = OS_ERROR;
//        pthread_exit(&ret_val);

	}
    /*Checking if the affinity applied is indeed effecive!*/
    ret = pthread_getaffinity_np(thread,sizeof(cpu_set_t),&cpuset);
    if (ret != 0)
    {
        printf("DATAGEN_THREAD[CPU-%d]## @@@@@@@@@@@@@@@@@@FAILED TO GET THREAD AFFINITY FOR OBSERVER THREAD@@@@@@@@@@@@@\n", threaddata);

//        ret_val = OS_ERROR;
//        pthread_exit(&ret_val);
    }

    printf("DATAGEN_THREAD[CPU-%d]## Set returned by pthread_getaffinity_np() contained:\n", threaddata);
    for (i = 0; i < __CPU_SETSIZE; i++)
    {
        if (CPU_ISSET(i, &cpuset))
        {
            printf("DATAGEN_THREAD[CPU-%d]##    CPU %d\n", threaddata, i);

        }
    }



    mqd = mq_open(mq_str, O_RDONLY);

    printf("DATAGEN_THREAD[CPU-%d]## MQ_OPEN returned %d\n", threaddata, mqd);
    if( mqd != (mqd_t)-1 )
    {
      printf("DATAGEN_THREAD[CPU-%d]## Message Queue Opened\n", threaddata);
    }
    else
    {
        printf("DATAGEN_THREAD[CPU-%d]## ERROR: Message queue %s open failed. Terminating... \n", threaddata, mq_str);

        ret_val = OS_ERROR;
        pthread_exit(&ret_val);
    }



    while (!finished)
    {

      printf("DATAGEN_THREAD[CPU-%d]##  Receiving message .... \n", threaddata);
        msg_len = mq_receive(mqd, msg_buff, MQ_MESSAGE_MAX_LENGTH, NULL);
        if(msg_len < 0)
        {
            printf("DATAGEN_THREAD[CPU-%d]##   Failed\n", threaddata);
        }
        else
        {
            msg_buff[MQ_MESSAGE_MAX_LENGTH-1] = '\0';
//          printf(" Successfully received %d bytes\n", (int)msg_len);
            printf("DATAGEN_THREAD[CPU-%d]## DG:: Message : %s \n", threaddata, msg_buff);
        }



        if (!strcmp(msg_buff,"INITIATE"))
        {
          printf("DATAGEN_THREAD[CPU-%d]## DG %d:: INITIATE\n", threaddata, dg_counter);
			dg_counter++;
			data_operate(PAYLOAD_SIZE_RD,threaddata);

        }
        else if (!strcmp(msg_buff,"TERMINATE"))
        {
            printf("DATAGEN_THREAD[CPU-%d]## DG:: TERMINATE\n", threaddata);
            finished = 1;
        }
        else if (!strcmp(msg_buff,"BARRIERS"))
        {
            printf("DATAGEN_THREAD[CPU-%d]## DG:: BARRIERS\n", threaddata);
            set_barriers();
        }
        else
        {
            printf("DATAGEN_THREAD[CPU-%d]## perf thread waiting for message\n", threaddata);
            /*poll after a delay*/
//          sleep(2);
        }


    }

    printf("DATAGEN_THREAD[CPU-%d]## Exiting from DG thread\n", threaddata);

    ret_val = SUCCESS;
	pthread_exit(&ret_val);

}


/*******************************************************************************
 * int launch_data_generator(void)                                             *
 *    - This launches a data generator threads that would perform some data    *
 *      write operations based on the commands of the IPC MsQ channel. One     *
 *      data generator thread per CPU in a selected cluster will be setup      *
 *    - arg1 - None                                                            *
 ******************************************************************************/
int launch_data_generator(void)
{
    pid_t pid;
	int ret_val = -1;
	int i = 0;
	int index = 0;

    REPORT_FN_DETAILS;

    /*Determine the data generator and data observer threads based on the cpu
      topology info discovered*/
    for (i=0; i<cpu_topology.cpu_count; i++)
    {
        /*Any cpu with is_readcpu set to False should have a data generator thread
          created*/
        if (!cpu_topology.cpu_info[i].is_readcpu)
        {
            /*This one means data generator thread needs to be created*/
            idwr[index].id = cpu_topology.cpu_info[i].cpu_id;
            sprintf(&(idwr[index].mq_name), "/DG%d_Q",index);
            printf("MAIN$$ Passed MQ %s to DG thread\n", &(idwr[index].mq_name));
            ret_val = pthread_create( &(data_gen_thread[index]), NULL, data_generator_main, (void*) &(idwr[index]));
            if (ret_val)
            {
                    printf("MAIN$$ ERROR: Unable to create data generator thread on CPU %d\n",idwr[index].id);
              	    return OS_ERROR;
            }
            index++;
        }
    }

	return SUCCESS;

}

/*~~~~~~~~~~~~ The above relates to data generator functions ~~~~~~~~~~~~~~~~~~*/

/*------ The following relates to data observer functions ---------------------*/

/*******************************************************************************
 * static void data_read(RD_T type, int cpu_num)                               *
 *    - This is function that does some data read operations from the shared   *
 *      memory                                                                 *
 *    - arg1 - type - the type of read operation                               *
 *    - arg2 - cpu_num - identifies the cpu where the data observer thread is  *
 *                  running                                                    *
 ******************************************************************************/

static void * data_read(RD_T type, int cpu_num)
{

    long long i = 0;
    long j = 0;
    long long sum = 0;
	int do_counter = 0;
	int temp = 0;
	int val = 0;
	long long payload_size = 0;
	long long inconsistent_op = 0;

    REPORT_FN_DETAILS;

	printf("DATAOBR_THREAD[cpu-%d]## rows = %ld cols = %ld \n",cpu_num, rows, cols);

        switch (type)
        {
            case PAYLOAD_SIZE_RD:
                printf("DATAOBR_THREAD[CPU-%d]## !!!!!!!Before PAYLOAD_SIZE_RD %lld !!!!!!!!!\n", cpu_num, pl_size);
				payload_size = pl_size;

                for (i = 0; i < payload_size; i++)
                {
                    sum += *((int*)array_2d+(cols*cci_mem_rdwr_payload[i].x_pos)+cci_mem_rdwr_payload[i].y_pos);
                    if (*((int*)array_2d+(cols*cci_mem_rdwr_payload[i].x_pos)+cci_mem_rdwr_payload[i].y_pos) == 255)
                    {
                        do_counter += 1;
                    }
				    else
				    {
				        inconsistent_op++;
				        if (0 == inconsistent_op%10000)
				        {
				            /*Print only when enough inconsistent reads are accumulated rather than swamping the console output*/
                            printf("DATAOBR_THREAD[CPU-%d]## !!!ERROR Cache inconsistent [%lld th time]!!! Read:: %d, Expected %d\n", cpu_num, inconsistent_op, val,255);
				        }

				    }
				}

                printf("DATAOBR_THREAD[CPU-%d]## !!!!!!!After PAYLOAD_SIZE_RD %lld !!!!!!!!!\n", cpu_num, pl_size);
    		break;

 		default:
			/*no op*/
			printf("DATAOBR_THREAD[CPU-%d]## This should have never happened\n", cpu_num);
			break;
	}

	printf("DATAOBR_THREAD[CPU-%d]## DO:: SUM = %lld \n", cpu_num, sum);
	printf("DATAOBR_THREAD[CPU-%d]## DO:: COUNT = %d \n", cpu_num, do_counter);



}

/*******************************************************************************
 * static void * data_observer_main(void)                                      *
 *    - This is the main entry function of the thread that performs the        *
 *      data read                                                              *
 *    - no arguments                                                           *
 ******************************************************************************/

static void * data_observer_main(void * ptr)
{
    mqd_t mqd;
    int ret = -1;
    int i = 0;
    int threaddata = 0;
    char msg_buff[MQ_MESSAGE_MAX_LENGTH];
    ssize_t msg_len;
    int ret_val = -1;
    int finished=0;
	static int do_counter=0;
	cpu_set_t cpuset;
	pthread_t thread;
 	THREADDATA_T * id;
 	char mq_str[16];

    REPORT_FN_DETAILS;

    id = (THREADDATA_T *)ptr;
    threaddata = id->id;
    strcpy(mq_str, id->mq_name);


    ((THREADDATA_T*)ptr)->threadid = syscall(SYS_gettid);

    printf("DATAOBR_THREAD[CPU-%d]## %d %s\n", threaddata, ((THREADDATA_T*)ptr)->id, ((THREADDATA_T*)ptr)->mq_name);

    printf ("DATAOBR_THREAD[CPU-%d]##  In do_threadid:: %d parent pid:: %d\n", threaddata, ((THREADDATA_T*)ptr)->threadid, getpid());

	CPU_ZERO(&cpuset);

	/*Set the affinity of data generator thread to CPUID passed via threaddata*/
	CPU_SET(threaddata,&cpuset);

	thread = pthread_self();

	ret = pthread_setaffinity_np(thread,sizeof(cpu_set_t),&cpuset);
	if (ret != 0)
	{
		printf("DATAOBR_THREAD[CPU-%d]## @@@@@@@@@@@@@@@@@@FAILED TO SET THREAD AFFINITY FOR OBSERVER THREAD@@@@@@@@@@@@@\n", threaddata);

//        ret_val = OS_ERROR;
//        pthread_exit(&ret_val);

	}
    /*Checking if the affinity applied is indeed effecive!*/
    ret = pthread_getaffinity_np(thread,sizeof(cpu_set_t),&cpuset);
    if (ret != 0)
    {
        printf("DATAOBR_THREAD[CPU-%d]## @@@@@@@@@@@@@@@@@@FAILED TO GET THREAD AFFINITY FOR OBSERVER THREAD@@@@@@@@@@@@@\n", threaddata);

//        ret_val = OS_ERROR;
//        pthread_exit(&ret_val);
    }

    printf("DATAOBR_THREAD[CPU-%d]## Set returned by pthread_getaffinity_np() contained:\n", threaddata);
    for (i = 0; i < __CPU_SETSIZE; i++)
    {
        if (CPU_ISSET(i, &cpuset))
        {
            printf("DATAOBR_THREAD[CPU-%d]##    CPU %d\n", threaddata, i);

        }
    }

    mqd = mq_open(mq_str, O_RDONLY);
//    mqd = mq_open("/DO1_Q", O_RDONLY);

    printf("DATAOBR_THREAD[CPU-%d]## MQ_OPEN returned %d\n", threaddata, mqd);
    if( mqd != (mqd_t)-1 )
    {
        printf("DATAOBR_THREAD[CPU-%d]## Message Queue Opened\n", threaddata);
    }
    else
    {
        printf("DATAOBR_THREAD[CPU-%d]## ERROR: Message queue %s open failed. Terminating... \n", threaddata, mq_str);

        ret_val = OS_ERROR;
        pthread_exit(&ret_val);
    }


    while (!finished)
    {

      printf("DATAOBR_THREAD[CPU-%d]## Receiving message .... \n", threaddata);
        msg_len = mq_receive(mqd, msg_buff, MQ_MESSAGE_MAX_LENGTH, NULL);
        if(msg_len < 0)
        {
            printf("DATAOBR_THREAD[CPU-%d]##   Failed\n", threaddata);
        }
        else
        {
            msg_buff[MQ_MESSAGE_MAX_LENGTH-1] = '\0';
//          printf(" Successfully received %d bytes\n", (int)msg_len);
            printf("DATAOBR_THREAD[CPU-%d]## DO:: Message : %s \n", threaddata, msg_buff);
        }



        if (!strcmp(msg_buff,"INITIATE"))
        {
          printf("DATAOBR_THREAD[CPU-%d]## DO %d:: INITIATE\n", threaddata, do_counter);
			do_counter++;
			data_read(PAYLOAD_SIZE_RD, threaddata);

        }
        else if (!strcmp(msg_buff,"TERMINATE"))
        {
            printf("DATAOBR_THREAD[CPU-%d]## DO:: TERMINATE\n", threaddata);
	printf("Hit any key: inside DO thread \n");
//	getchar();
            finished = 1;

        }
        else if (!strcmp(msg_buff,"BARRIERS"))
        {
            printf("DATAOBR_THREAD[CPU-%d]## DO:: BARRIERS\n", threaddata);
            set_barriers();
        }
        else
        {
            printf("DATAOBR_THREAD[CPU-%d]## perf thread waiting for message\n", threaddata);
            /*poll after a delay*/
//          sleep(2);
        }


    }

    printf("DATAOBR_THREAD[CPU-%d]## Exiting from DO thread\n", threaddata);

	printf("Hit any key: before pthread_exit \n");
//	getchar();
    ret_val = SUCCESS;
	pthread_exit(&ret_val);

}

/*******************************************************************************
 * int launch_data_observer(void)                                              *
 *    - This launches a data observer threads that would perform some data     *
 *      read operations based on the commands of the IPC MsQ channel. One      *
 *      data observer thread per CPU in a selected cluster will be setup       *
 *    - arg1 - None                                                            *
 ******************************************************************************/
int launch_data_observer(void)
{
    pid_t pid;
	int ret_val = -1;
	int i = 0;
	int index = 0;

    REPORT_FN_DETAILS;

    /*Determine the data generator and data observer threads based on the cpu
      topology info discovered*/
    for (i=0; i<cpu_topology.cpu_count; i++)
    {
        /*Any cpu with is_readcpu set to True should have a data observer thread
          created*/
        if (cpu_topology.cpu_info[i].is_readcpu)
        {
            idrd[index].id = cpu_topology.cpu_info[i].cpu_id;
            sprintf(&(idrd[index].mq_name), "/DO%d_Q",index);
            printf("Passed MQ %s to DO thread\n", &(idrd[index].mq_name));
            ret_val = pthread_create( &(data_obsr_thread[index]), NULL, data_observer_main, (void*) &(idrd[index]));
            if (ret_val)
            {
                    printf("ERROR: Unable to create data observer thread on CPU %d\n",idrd[i].id);
              	    return OS_ERROR;
            }
            index++;
        }

    }

	return SUCCESS;

}

/*~~~~~~~~~~~~ The above relates to data observer functions ~~~~~~~~~~~~~~~~~~~*/


/*******************************************************************************
 * int launch_ccitest_payload_operations(void)                                 *
 *    - This function initialises the ccitest setup by allocating the needed   *
 *      memory and then launching the data generator and data observer threads *
 *      to act on the command received from the search algorithm               *
 *    - arg1 - None                                                            *
 ******************************************************************************/
int launch_ccitest_payload_operations(void)
{

	int ret_val = -1;
	int i = 0;
	char mq_str[16];
    mqd_t mqd;

	REPORT_FN_DETAILS;

	/*First initialise the heap memory*/
	ret_val = initialise_memory_pattern();
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to initialise heap memory for CCITest Payload\n");
	    return OS_ERROR;
	}

    /*First extract the info about topology to determine the number of data
      generator and data observer threads. Also determine how the affinity
      needs to be applied*/
    ret_val = discover_topology_info();
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR: Discovery of the topology failed\n");
        return OS_ERROR;
    }

    ret_val = display_topology_info();
    if (SUCCESS != ret_val)
    {
        printf("MAIN$$ ERROR: Display of the topology failed\n");
        return OS_ERROR;
    }

    /*Determine the data generator and data observer thread counts before
      setting up the message queues*/
    for (i=0; i<cpu_topology.cpu_count; i++)
    {
        if (cpu_topology.cpu_info[i].is_readcpu)
        {
            /*Any cpu with is_readcpu set to True increment the dg count*/
            do_thread_count++;
        }
        else
        {
            /*Any cpu with is_readcpu set to False increment the dg count*/
            dg_thread_count++;
        }
    }

    /*Setup the DG MsgQ for IPC starting with a cleanup*/
    /*Cleanup anything that might be in stale state!*/
    for (i=0; i<MAX_CPU_COUNT; i++)
    {
        sprintf(mq_str, "/DG%d_Q",i);
        ret_val = mq_unlink(mq_str);
        /*You don;t want to check the ret_val here as this is a clean-up step*/
        mqd_wr[i] =  0;

    }


    printf("MAIN$$ dg_thread_count = %d\n",dg_thread_count);

    for (i=0; i<dg_thread_count; i++)
    {
        /*Now create the Msg Q for Data generator threads*/
        sprintf(mq_str, "/DG%d_Q",i);
        mqd = mq_open(mq_str, O_WRONLY | O_CREAT, MQ_MODE, &attr);

        if( mqd != (mqd_t)-1 )
        {
            mqd_wr[i] = mqd;
            printf("MAIN$$ Message Queue Opened for [%s] with [%d] \n", mq_str, mqd_wr[i]);
        }
        else
        {
            printf("MAIN$$ ERROR: Message queue %s open failed. Terminating... \n", mq_str);
            return OS_ERROR;
        }
    }

	/*First launch the data generator threads*/
	ret_val = launch_data_generator();
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to launch data generator thread(s) for CCITest Payload\n");
	    return OS_ERROR;
	}


    /*Setup the DO MsgQ for IPC starting with a cleanup*/
    /*Cleanup anything that might be in stale state!*/
    for (i=0; i<MAX_CPU_COUNT; i++)
    {
        sprintf(mq_str, "/DO%d_Q",i);
        ret_val = mq_unlink(mq_str);
        /*You don;t want to check the ret_val here as this is a clean-up step*/
        mqd_rd[i] =  0;

    }

    printf("MAIN$$ do_thread_count = %d\n",do_thread_count);

    for (i=0; i<do_thread_count; i++)
    {
        /*Now create the Msg Q for Data observer threads*/
        sprintf(mq_str, "/DO%d_Q",i);
        mqd = mq_open(mq_str, O_WRONLY | O_CREAT, MQ_MODE, &attr);
        if( mqd != (mqd_t)-1 )
        {
            mqd_rd[i] = mqd;
            printf("MAIN$$ Message Queue Opened for [%s] with [%d]\n", mq_str, mqd_rd[i]);
        }
        else
        {
            printf("MAIN$$ ERROR: Message queue %s open failed. Terminating... \n", mq_str);
            return OS_ERROR;
        }
    }

	/*Then launch the data observer threads*/
	ret_val = launch_data_observer();
    if (SUCCESS != ret_val)
	{
		printf("MAIN$$ ERROR: Unable to launch data observer thread(s) for CCITest Payload\n");
	    return OS_ERROR;
	}


    for (i=0; i<MAX_CPU_COUNT; i++)
    {
        printf("MAIN$$ mqd_wr[%d] = [%d] || mqd_rd[%d] = [%d]\n", i, mqd_wr[i], i, mqd_rd[i]);
    }




	return SUCCESS;

}

/*******************************************************************************
 * int shutdown_ccitest_payload_operations(void)                               *
 *    - This function shutdown the ccitest setup by tearing down the initiated *
 *      data generator and observer threads                                    *
 *    - arg1 - None                                                            *
 ******************************************************************************/
int shutdown_ccitest_payload_operations(void)
{

	int ret_val = -1;
	int i = 0;
	char mq_str[16];
	int* thread_return_wr[1];
	int* thread_return_rd[1];

	REPORT_FN_DETAILS;

    for (i=0; i<MAX_CPU_COUNT; i++)
    {
        printf("MAIN$$ mqd_wr[%d] = [%d] || mqd_rd[%d] = [%d]\n", i, mqd_wr[i], i, mqd_rd[i]);
    }



    /*Tear down the data generator threads first*/
    for (i=0; i<dg_thread_count; i++)
    {
        /*Now send the terminate signal*/
        sprintf(mq_str, "/DG%d_Q",i);

        printf("MAIN$$ Sending message to [%s] with mqd_wr id [%d].... ",mq_str,mqd_wr[i]);
        ret_val = mq_send(mqd_wr[i], "TERMINATE", 10, MQ_MESSAGE_PRIORITY);
        if(ret_val)
        {
            printf("MAIN$$ TERMINATE Failed\n");

        }
        else
        {
            printf("MAIN$$ TERMINATE Done\n");

//	printf("Hit any key: \n");
//	getchar();

        /*Wait for the data generator thread before terminating*/
        printf("MAIN$$ data_gen_thread[%d] = %d\n", i, data_gen_thread[i]);
        pthread_join(data_gen_thread[i],(void**)&(thread_return_wr[0]));
        printf("MAIN$$ Data Generator thread returned %d\n",*(thread_return_wr[0]));

        }

//        sleep(10);

        /* Now close the message queue */
        ret_val = mq_close(mqd_wr[i]);
        if(ret_val)
            printf("MAIN$$ Message queue close failed\n");
        else
            printf("MAIN$$ Message Queue Closed\n");

        ret_val = mq_unlink(mq_str);
        if(ret_val)
            printf("MAIN$$ Message queue unlink failed\n");
        else
            printf("MAIN$$ Message Queue unlinked\n");
    }

#if 1
    /*Tear down the data observer threads first*/
    for (i=0; i<do_thread_count; i++)
    {
        /*Now send the terminate signal*/
        sprintf(mq_str, "/DO%d_Q",i);

        printf("MAIN$$ Sending message to [%s] with mqd_rd id [%d].... ",mq_str,mqd_rd[i]);
        ret_val = mq_send(mqd_rd[i], "TERMINATE", 10, MQ_MESSAGE_PRIORITY);
        if(ret_val)
        {
            printf("MAIN$$ TERMINATE Failed\n");
        }
        else
        {
            printf("MAIN$$ TERMINATE Done\n");

        /*Wait for the data observer thread before terminating*/
        printf("MAIN$$ data_obsr_thread[%d] = %d\n", i, data_obsr_thread[i]);
//        pthread_join(data_obsr_thread[i],(void**)&(thread_return_rd[0]));
        printf("MAIN$$ Data Observer thread returned %d\n",*(thread_return_rd[0]));

        }

//        sleep(10);

        /* Now close the message queue */
        ret_val = mq_close(mqd_rd[i]);
        if(ret_val)
            printf("MAIN$$ Message queue close failed\n");
        else
            printf("MAIN$$ Message Queue Closed\n");

        ret_val = mq_unlink(mq_str);
        if(ret_val)
            printf("MAIN$$ Message queue unlink failed\n");
        else
            printf("MAIN$$ Message Queue unlinked\n");
    }

#endif

	return SUCCESS;

}
